-- $Id: load.lua 2130 2016-01-23 22:22:11Z dsb $
-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham ("Zizzo")
-- dsb-tome@killerbunnies.org

local TimedEffects = require 'engine.interface.ActorTemporaryEffects'

if config.settings.tome.loading_zomnibus then
  TimedEffects:loadDefinition('/data-zomnibus/landmark_timed_effects.lua')
else
  TimedEffects:loadDefinition('/data-landmark/timed_effects.lua')
end


local function handle_landmark(item)
  local Dialog = require 'engine.ui.Dialog'

  if not item then
    print('[LANDMARK] handle_landmark:  ??? no item')
    return
  end
  if item.x < 0 or item.x >= game.level.map.w or item.y < 0 or item.y >= game.level.map.h then
    print(('[LANDMARK] handle_landmark(%s "%s" @%d,%d):  ??? dest out of bounds [%dx%d]'):format(item.type, item.name, item.x, item.y, game.level.map.w, game.level.map.h))
    return
  end
  print(('[LANDMARK] handle_landmark(%s "%s" @%d,%d)'):format(item.type, item.name, item.x, item.y))

  local list = {
    { name = 'Go to ' .. item.name, go_to = true },
    { name = 'View ' .. item.name .. ' on map', go_to = false },
  }

  local d
  local function cb(l)
    if l and l.go_to then
      print('[LANDMARK] handle_landmark:  selected go-to')
      local p = game.player
      -- Dodgy Hack(TM):  Remember what landmark we've chosen for the
      -- benefit of our hijacked Player:autoExplore().
      p:setEffect(p.EFF_GOING_TO_LANDMARK, 1, { target_name=item.short_name, target_x = item.x, target_y = item.y })
      -- Convoluted Hack(TM):  mouseMove() apparently doesn't work while
      -- a dialog is up, so we arrange for it to be called after we close
      -- this dialog.
      print('[LANDMARK] handle_landmark:  installing unload() method')
      d.unload = function(self)
	print('[LANDMARK] handle_landmark:  in unload() method, installing onTickEnd() callback')
	game:onTickEnd(function()
	  print(('[LANDMARK] handle_landmark:  in onTickEnd() callback, going to %d,%d'):format(item.x, item.y))
	  game.player:mouseMove(item.x, item.y, nil, nil, true)
	end)
      end
    elseif l then
      print('[LANDMARK] handle_landmark:  selected view')
      game.level.map:moveViewSurround(item.x, item.y, 1000, 1000)
    end
  end
  d = Dialog:listPopup('Landmark:  ' .. item.name, 'Action:', list, 800, 100, cb)
end

local landmark_types = {
  town		= { 1, 'Towns' },
  zone		= { 2, 'Zones' },
  store		= { 3, 'Stores', by_dist=true },
  lev_change	= { 4, 'Level exits' },
  npcs		= { 5, 'People', by_dist=true },
  misc		= { 6, 'Miscellaneous', by_dist=true },
}
local landmark_npcs = {
  ['Agrimley the Hermit']			= true,
  ['Sun Paladin Guren']				= true,
  ['Linaniil, Supreme Archmage of Angolwen']	= true,
  ['Shady cornac man']				= true,
  ['Protector Myssil']				= true,
  ['High Sun Paladin Aeryn']			= true,
}

local function grid_landmark_type(g)
  if not g then return nil end
  if g.change_zone then
    return g.change_zone:find('^town%-') and 'town' or 'zone'
  end
  -- Slight Hack(TM) for Slazish Fens exit portal.
  if g.name == 'coral portal' then return 'zone' end
  if g.change_level then return 'lev_change' end
  if g.special and not g.autoexplore_ignore then return 'misc' end
  if g.orb_portal or g.name == 'font of life' then return 'misc' end
  return nil
end

local function landmark_entry(e, type, x, y)
  local dist = core.fov.distance(x, y, game.player.x, game.player.y, true)
  local name = ('%s #YELLOW#(%d,%d, est.dist. %.2f)#LAST#'):format(e.name, x, y, dist)
  local lev = nil
  if type == 'zone' or type == 'town' then
    lev = 999
    -- We loosely mimic mod.class.Game:tooltip(), which notes: "Lets make
    -- very very sure that funky weird zone files dont explode things"
    local ok, z = pcall(function()
      local Zone = require 'engine.Zone'
      local fakezone = {short_name=e.change_zone}
      local base = Zone.getBaseName(fakezone)
      local f = loadfile(base.."/zone.lua")
      if f then
	setfenv(f, setmetatable({self=fakezone, short_name=fakezone.short_name}, {__index=_G}))
	local ok, zz = pcall(f)
	return zz
      end
    end)
    if ok and z and z.level_range then
      lev = z.level_range[1]
      local plev = game.player.level
      local color = plev <= lev - 10 and 'CRIMSON' or plev <= lev - 4 and 'ORANGE' or 'YELLOW'
      name = name .. (' #%s#(min.level: %d)'):format(color, lev)
    end
    -- Careful; Slazish Fens exit portal is a 'zone'-type landmark but
    -- doesn't have a 'change_zone' field.
    if game.level.entrance_glow and e.change_zone and not game.visited_zones[e.change_zone] then
      name = name .. ' #LIGHT_BLUE#[never visited]#LAST#'
    end
  end
  if type == 'lev_change' then lev = e.change_level end
  return { name=name, short_name=e.name, type=type, x=x, y=y, dist=dist, lev=lev }
end

local function landmark_order(l1, l2)
  local t1, t2 = landmark_types[l1.type], landmark_types[l2.type]
  -- First sort by landmark type order.  Sort unrecognized landmark types
  -- to the end, and there by type name.
  if l1.type ~= l2.type then
    if t1 and t2 then
      return t1[1] < t2[1]
    elseif t1 or t2 then
      return t1 and false or true
    else
      return l1.type < l2.type
    end
  end
  -- Within a landmark type, sort by level if both define one.
  if l1.lev and l2.lev then
    if l1.lev ~= l2.lev then return l1.lev < l2.lev end
  elseif l1.lev or l2.lev then
    return l1.lev and false or true
  end
  -- Sort miscellaneous landmarks and stores by distance from the player.
  if t1.by_dist and l1.dist ~= l2.dist then
    return l1.dist < l2.dist
  end
  -- Failing all else, sort by name.
  return l1.name < l2.name
end

local function show_landmark_list()
  local Map = require 'engine.Map'
  local Dialog = require 'engine.ui.Dialog'
  local list = {}

  for x = 0, game.level.map.w - 1 do
    for y = 0, game.level.map.h - 1 do
      if game.level.map.has_seens(x, y) then
	local g = game.level.map(x, y, Map.TERRAIN)
	local type = grid_landmark_type(g)
	if type then list[#list+1] = landmark_entry(g, type, x, y) end
	local a = game.level.map(x, y, Map.ACTOR)
	if a and a.name and landmark_npcs[a.name] then
	  list[#list+1] = landmark_entry(a, 'npcs', x, y)
	end
	local t = game.level.map(x, y, Map.TRAP)
	if t and t.is_store then
	  list[#list+1] = landmark_entry(t, 'store', x, y)
	end
      end
    end
  end

  table.sort(list, landmark_order)
  -- Loop over the list, inserting subtitle entries.
  local i = 1
  while i <= #list do
    if i == 1 or list[i].type ~= list[i-1].type then
      local name = landmark_types[list[i].type] and landmark_types[list[i].type][2] or '??? ('..list[i].type..')'
      table.insert(list, i, {
	name = '#GREY##{bold}#'..name..':#{normal}##LAST#',
	subtitle = true,
      })
      i = i + 2
    else
      i = i + 1
    end
  end

  local zone_name
  if game.zone.display_name then
    zone_name = game.zone.display_name()
  else
    local lev = game.level.level
    if game.level.data.reverse_level_display then lev = 1 + game.level.data.max_level - lev end
    if game.zone.max_level == 1 then
      zone_name = game.zone.name
    else
      zone_name = ("%s (%d)"):format(game.zone.name, lev)
    end
  end

  -- We unroll Dialog:listPopup() so that we can prevent its ACCEPT key
  -- binding from closing the dialog on a subtitle line.
  local d = Dialog.new('Landmarks - '..zone_name, 1, 1)
  local desc = require('engine.ui.Textzone').new {
    width = 600,
    auto_height = true,
    text = 'Select landmark to view or travel to:',
  }
  local h = math.max(600, game.h * 0.8)
  local l = require('engine.ui.List').new {
    width = 600,
    height = h - 16 - desc.h,
    list = list,
    fct = function() d.key:triggerVirtual('ACCEPT') end,
    scrollbar = true,
  }
  local sep = require('engine.ui.Separator').new { dir='vertical', size=588 }
  d:loadUI {
    { left=3, top=3, ui=desc },
    { left=3, top=6+desc.h, ui=sep },
    { left=3, bottom=3, ui=l },
  }
  d.key:addBind('EXIT', function() game:unregisterDialog(d) end)
  d.key:addBind('ACCEPT', function()
    if list[l.sel].subtitle then return end
    game:unregisterDialog(d)
    handle_landmark(list[l.sel])
  end)
  d:setFocus(l)
  d:setupUI(true, true)
  game:registerDialog(d)
end

class:bindHook('Game:alterGameMenu', function(self, data)
  local cb = function()
    data.unregister()
    if game and game.level and game.level.map then show_landmark_list() end
  end
  -- Insert our menu item after 'Show ingredients'.
  local idx = math.min(#data.menu, 5)
  local present = false
  for i, item in ipairs(data.menu) do
    if type(item) == 'table' then
      if item[1] == 'Show ingredients' then idx = i+1 end
      if item[1] == 'Show known Landmarks' then present = true end
    end
  end
  if not present then
    table.insert(data.menu, idx, { 'Show known Landmarks', cb })
  end
end)
